@@ -5,12 +5,13 @@ from ctypes import CDLL, POINTER, c_double, c_char_p, pointer
 from ctypes.util import find_library
 import fnmatch
 from collections import MutableMapping
-
+from datetime import datetime, timedelta
 import redis
+from redis.exceptions import ResponseError
 import redis.client
 
 
-__version__ = '0.3.1'
+__version__ = '0.4.0'
 DATABASES = {}
 
 _libc = CDLL(find_library('c'))
@@ -22,8 +23,10 @@ _strtod = _libc.strtod
 class _StrKeyDict(MutableMapping):
     def __init__(self, *args, **kwargs):
         self._dict = dict(*args, **kwargs)
+        self._ex_keys = {}
 
     def __getitem__(self, key):
+        self._update_expired_keys()
         return self._dict[str(key)]
 
     def __setitem__(self, key, value):
@@ -38,6 +41,29 @@ class _StrKeyDict(MutableMapping):
     def __iter__(self):
         return iter(self._dict)
 
+    def expire(self, key, timestamp):
+        self._ex_keys[key] = timestamp
+
+    def expiring(self, key):
+        if not key in self._ex_keys.keys():
+            return None
+        return self._ex_keys[key]
+
+    def _update_expired_keys(self):
+        now = datetime.now()
+        deleted = []
+        for key in self._ex_keys.keys():
+            if now > self._ex_keys[key]:
+                deleted.append(key)
+
+        for key in deleted:
+            del self._ex_keys[key]
+            del self[key]
+
+    def clear(self):
+        super(_StrKeyDict, self).clear()
+        self._ex_keys.clear()
+
 
 class FakeStrictRedis(object):
     def __init__(self, db=0, **kwargs):
@@ -46,6 +72,7 @@ class FakeStrictRedis(object):
         self._db = DATABASES[db]
         self._db_num = db
 
+
     def flushdb(self):
         DATABASES[self._db_num].clear()
         return True
@@ -59,6 +86,17 @@ class FakeStrictRedis(object):
         self._db[key] += value
         return len(self._db[key])
 
+    def bitcount(self, name, start=0, end=-1):
+        if end == -1:
+            end = None
+        else:
+            end += 1
+        try:
+            s = self._db[name][start:end]
+            return sum([bin(ord(l)).count('1') for l in s])
+        except KeyError:
+            return 0
+
     def decr(self, name, amount=1):
         try:
             self._db[name] = self._db.get(name, 0) - amount
@@ -72,10 +110,18 @@ class FakeStrictRedis(object):
     __contains__ = exists
 
     def expire(self, name, time):
-        pass
+        if self.exists(name):
+            self._db.expire(name, datetime.now() + timedelta(seconds=time))
+        else:
+            return False
 
     def expireat(self, name, when):
-        pass
+        if not self.exists(name):
+            return False
+        if isinstance(when, datetime):
+            self._db.expire(name, when)
+        else:
+            self._db.expire(name, datetime.fromtimestamp(when))
 
     def get(self, name):
         value = self._db.get(name)
@@ -86,7 +132,7 @@ class FakeStrictRedis(object):
         return self._db[name]
 
     def getbit(self, name, offset):
-        "Returns a boolean indicating the value of ``offset`` in ``name``"
+        """Returns a boolean indicating the value of ``offset`` in ``name``"""
         val = self._db.get(name, '\x00')
         byte = offset / 8
         remaining = offset % 8
@@ -173,9 +219,19 @@ class FakeStrictRedis(object):
         else:
             return self.rename(src, dst)
 
-    def set(self, name, value):
-        self._db[name] = value
-        return True
+    def set(self, name, value, ex=None, px=None, nx=False, xx=False):
+        if (not nx and not xx) \
+        or (nx and self._db.get(name, None) is None) \
+        or (xx and not self._db.get(name, None) is None):
+            if ex > 0:
+                self._db.expire(name, datetime.now() + timedelta(seconds=ex))
+            elif px > 0:
+                self._db.expire(name, datetime.now() + timedelta(milliseconds=px))
+            self._db[name] = value
+            return True
+        else:
+            return None
+
     __setitem__ = set
 
     def setbit(self, name, offset, value):
@@ -197,14 +253,19 @@ class FakeStrictRedis(object):
         self._db[name] = ''.join(reconstructed)
 
     def setex(self, name, time, value):
-        return self.set(name, value)
+        return self.set(name, value, ex=time)
+
+    def psetex(self, name, time_ms, value):
+        if time_ms == 0:
+            raise ResponseError("invalid expire time in SETEX")
+        return self.set(name, value, px=time_ms)
 
     def setnx(self, name, value):
-        if name in self._db:
+        result = self.set(name, value, nx=True)
+        # Real Redis returns False from setnx, but None from set(nx=...)
+        if not result:
             return False
-        else:
-            self._db[name] = value
-            return True
+        return result
 
     def setrange(self, name, offset, value):
         pass
@@ -229,7 +290,20 @@ class FakeStrictRedis(object):
     getrange = substr
 
     def ttl(self, name):
-        pass
+        if name not in self._db:
+            return None
+
+        exp_time = self._db.expiring(name)
+        if not exp_time:
+            return None
+
+        now = datetime.now()
+        if now > exp_time:
+            return None
+        else:
+            return round((exp_time - now).days * 3600 * 24
+                         + (exp_time - now).seconds
+                         + (exp_time - now).microseconds / 1E6)
 
     def type(self, name):
         pass
@@ -241,14 +315,14 @@ class FakeStrictRedis(object):
         pass
 
     def delete(self, *names):
-        any_deleted = False
+        deleted = 0
         for name in names:
             try:
                 del self._db[name]
-                any_deleted = True
+                deleted += 1
             except KeyError:
                 continue
-        return any_deleted
+        return deleted
 
     def sort(self, name, start=None, num=None, by=None, get=None, desc=False,
              alpha=False, store=None):
@@ -483,7 +557,7 @@ class FakeStrictRedis(object):
             if k in h:
                 del h[k]
                 rem += 1
-        return rem > 0
+        return rem
 
     def hexists(self, name, key):
         "Returns a boolean indicating if ``key`` exists within hash ``name``"
@@ -543,10 +617,11 @@ class FakeStrictRedis(object):
         self._db.setdefault(name, {}).update(mapping)
         return True
 
-    def hmget(self, name, keys):
+    def hmget(self, name, keys, *args):
         "Returns a list of values ordered identically to ``keys``"
         h = self._db.get(name, {})
-        return [h.get(k) for k in keys]
+        all_keys = self._list_or_args(keys, args)
+        return [h.get(k) for k in all_keys]
 
     def hvals(self, name):
         "Return the list of values within hash ``name``"
@@ -565,7 +640,7 @@ class FakeStrictRedis(object):
 
     def sdiff(self, keys, *args):
         "Return the difference of sets specified by ``keys``"
-        all_keys = redis.client.list_or_args(keys, args)
+        all_keys = self._list_or_args(keys, args)
         diff = self._db.get(all_keys[0], set()).copy()
         for key in all_keys[1:]:
             diff -= self._db.get(key, set())
@@ -582,7 +657,7 @@ class FakeStrictRedis(object):
 
     def sinter(self, keys, *args):
         "Return the intersection of sets specified by ``keys``"
-        all_keys = redis.client.list_or_args(keys, args)
+        all_keys = self._list_or_args(keys, args)
         intersect = self._db.get(all_keys[0], set()).copy()
         for key in all_keys[1:]:
             intersect.intersection_update(self._db.get(key, set()))
@@ -632,11 +707,11 @@ class FakeStrictRedis(object):
         a_set = self._db.setdefault(name, set())
         card = len(a_set)
         a_set -= set(values)
-        return (card - len(a_set)) > 0
+        return card - len(a_set)
 
     def sunion(self, keys, *args):
         "Return the union of sets specifiued by ``keys``"
-        all_keys = redis.client.list_or_args(keys, args)
+        all_keys = self._list_or_args(keys, args)
         union = self._db.get(all_keys[0], set()).copy()
         for key in all_keys[1:]:
             union.update(self._db.get(key, set()))
@@ -805,7 +880,7 @@ class FakeStrictRedis(object):
             if v in z:
                 del z[v]
                 rem += 1
-        return rem > 0
+        return rem
 
     def zremrangebyrank(self, name, min, max):
         """
